@@ -16,27 +16,40 @@ from typing import (
 )
 
 import anyio
-from sqlalchemy import Column, asc, desc, func, inspect, or_, select
-from sqlalchemy.engine.base import Engine
+from sqlalchemy import Column, asc, desc, func, inspect, or_
 from sqlalchemy.exc import NoInspectionAvailable
-from sqlalchemy.ext.asyncio import AsyncEngine
 from sqlalchemy.orm import (
     ColumnProperty,
     RelationshipProperty,
-    selectinload,
+    joinedload,
     sessionmaker,
 )
 from sqlalchemy.orm.attributes import InstrumentedAttribute
 from sqlalchemy.sql.elements import ClauseElement
+from sqlalchemy.sql.expression import Select, select
 from starlette.requests import Request
 from starlette.responses import StreamingResponse
+from starlette.templating import Jinja2Templates
 from wtforms import Field, Form
 
+from sqladmin._queries import (
+    delete_async,
+    delete_sync,
+    insert_async,
+    insert_sync,
+    update_async,
+    update_sync,
+)
+from sqladmin._types import ENGINE_TYPE, MODEL_ATTR_TYPE
 from sqladmin.exceptions import InvalidColumnError, InvalidModelError
+from sqladmin.formatters import BASE_FORMATTERS
 from sqladmin.forms import get_model_form
 from sqladmin.helpers import (
     Writer,
-    as_str,
+    get_attributes,
+    get_column_python_type,
+    get_primary_key,
+    get_relationships,
     prettify_class_name,
     secure_filename,
     slugify_class_name,
@@ -45,12 +58,14 @@ from sqladmin.helpers import (
 from sqladmin.pagination import Pagination
 
 __all__ = [
-    "ModelAdmin",
+    "BaseView",
+    "ModelView",
+    "ModelView",
 ]
 
 
-class ModelAdminMeta(type):
-    """Metaclass used to specify class variables in ModelAdmin.
+class ModelViewMeta(type):
+    """Metaclass used to specify class variables in ModelView.
 
     Danger:
         This class should almost never be used directly.
@@ -58,7 +73,7 @@ class ModelAdminMeta(type):
 
     @no_type_check
     def __new__(mcls, name, bases, attrs: dict, **kwargs: Any):
-        cls: Type["ModelAdmin"] = super().__new__(mcls, name, bases, attrs)
+        cls: Type["ModelView"] = super().__new__(mcls, name, bases, attrs)
 
         model = kwargs.get("model")
 
@@ -66,15 +81,13 @@ class ModelAdminMeta(type):
             return cls
 
         try:
-            mapper = inspect(model)
+            inspect(model)
         except NoInspectionAvailable:
             raise InvalidModelError(
                 f"Class {model.__name__} is not a SQLAlchemy model."
             )
 
-        assert len(mapper.primary_key) == 1, "Multiple PK columns not supported."
-
-        cls.pk_column = mapper.primary_key[0]
+        cls.pk_column = get_primary_key(model)
         cls.identity = slugify_class_name(model.__name__)
         cls.model = model
 
@@ -82,6 +95,9 @@ class ModelAdminMeta(type):
         cls.name_plural = attrs.get("name_plural", f"{cls.name}s")
         cls.icon = attrs.get("icon")
 
+        cls.list_query = attrs.get("list_query", select(model))
+        cls.count_query = attrs.get("count_query", select(func.count(cls.pk_column)))
+
         mcls._check_conflicting_options(["column_list", "column_exclude_list"], attrs)
         mcls._check_conflicting_options(
             ["form_columns", "form_excluded_columns"], attrs
@@ -101,7 +117,7 @@ class ModelAdminMeta(type):
             raise AssertionError(f"Cannot use {' and '.join(keys)} together.")
 
 
-class BaseModelAdmin:
+class BaseModelView:
     def is_visible(self, request: Request) -> bool:
         """Override this method if you want dynamically
         hide or show administrative views from SQLAdmin menu structure
@@ -119,16 +135,65 @@ class BaseModelAdmin:
         return True
 
 
-class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
+class BaseView(BaseModelView):
+    """Base class for defining admnistrative views for the model.
+
+    ???+ usage
+        ```python
+        from sqladmin import BaseView, expose
+
+        class CustomAdmin(BaseView):
+            name = "Custom Page"
+            icon = "fa-solid fa-chart-line"
+
+            @expose("/custom", methods=["GET"])
+            def test_page(self, request: Request):
+                return self.templates.TemplateResponse(
+                    "custom.html",
+                    context={"request": request},
+                )
+
+        admin.add_base_view(CustomAdmin)
+        ```
+    """
+
+    # Internals
+    is_model: ClassVar[bool] = False
+    url_path_for: ClassVar[Callable]
+    templates: ClassVar[Jinja2Templates]
+
+    name: ClassVar[str] = ""
+    """Name of the view to be displayed."""
+
+    identity: ClassVar[str] = ""
+    """Same as name but it will be used for URL of the endpoints."""
+
+    methods: ClassVar[List[str]] = ["GET"]
+    """List of method names for the endpoint.
+    By default it's set to `["GET"]` only.
+    """
+
+    include_in_schema: ClassVar[bool] = True
+    """Control whether this endpoint
+    should be included in the schema.
+    """
+
+    icon: ClassVar[str]
+    """Display icon for ModelAdmin in the sidebar.
+    Currently only supports FontAwesome icons.
+    """
+
+
+class ModelView(BaseView, metaclass=ModelViewMeta):
     """Base class for defining admnistrative behaviour for the model.
 
     ???+ usage
         ```python
-        from sqladmin import ModelAdmin
+        from sqladmin import ModelView
 
         from mymodels import User # SQLAlchemy model
 
-        class UserAdmin(ModelAdmin, model=User):
+        class UserAdmin(ModelView, model=User):
             can_create = True
         ```
     """
@@ -137,34 +202,16 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
 
     # Internals
     pk_column: ClassVar[Column]
-    identity: ClassVar[str]
     sessionmaker: ClassVar[sessionmaker]
-    engine: ClassVar[Union[Engine, AsyncEngine]]
+    engine: ClassVar[ENGINE_TYPE]
     async_engine: ClassVar[bool]
-    url_path_for: ClassVar[Callable]
-
-    # Metadata
-    name: ClassVar[str] = ""
-    """Name of ModelAdmin to display.
-    Default value is set to Model class name.
-    """
+    is_model: ClassVar[bool] = True
 
     name_plural: ClassVar[str] = ""
-    """Plural name of ModelAdmin.
+    """Plural name of ModelView.
     Default value is Model class name + `s`.
     """
 
-    icon: ClassVar[str] = ""
-    """Display icon for ModelAdmin in the sidebar.
-    Currently only supports FontAwesome icons.
-
-    ???+ example
-        ```python
-        class UserAdmin(ModelAdmin, model=User):
-            icon = "fa-solid fa-user"
-        ```
-    """
-
     # Permissions
     can_create: ClassVar[bool] = True
     """Permission for creating new Models. Default value is set to `True`."""
@@ -195,7 +242,7 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
 
     ???+ example
         ```python
-        class UserAdmin(ModelAdmin, model=User):
+        class UserAdmin(ModelView, model=User):
             column_list = [User.id, User.name]
         ```
     """
@@ -206,7 +253,7 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
 
     ???+ example
         ```python
-        class UserAdmin(ModelAdmin, model=User):
+        class UserAdmin(ModelView, model=User):
             column_exclude_list = [User.id, User.name]
         ```
     """
@@ -219,7 +266,7 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
 
     ???+ example
         ```python
-        class UserAdmin(ModelAdmin, model=User):
+        class UserAdmin(ModelView, model=User):
             column_formatters = {User.name: lambda m, a: m.name[:10]}
         ```
 
@@ -228,7 +275,7 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
         ```python
         def formatter(model, attribute):
             # `model` is model instance
-            # `attribute` is a Union[Column, ColumnProperty, RelationshipProperty]
+            # `attribute` is a Union[ColumnProperty, RelationshipProperty]
             pass
         ```
     """
@@ -239,7 +286,7 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
 
     ???+ example
         ```python
-        class UserAdmin(ModelAdmin, model=User):
+        class UserAdmin(ModelView, model=User):
             page_size = 25
         ```
     """
@@ -250,7 +297,7 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
 
     ???+ example
         ```python
-        class UserAdmin(ModelAdmin, model=User):
+        class UserAdmin(ModelView, model=User):
             page_size_options = [50, 100]
         ```
     """
@@ -262,7 +309,7 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
 
     ???+ example
         ```python
-        class UserAdmin(ModelAdmin, model=User):
+        class UserAdmin(ModelView, model=User):
             column_searchable_list = [User.name]
         ```
     """
@@ -272,11 +319,38 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
 
     ???+ example
         ```python
-        class UserAdmin(ModelAdmin, model=User):
+        class UserAdmin(ModelView, model=User):
             column_sortable_list = [User.name]
         ```
     """
 
+    column_default_sort: ClassVar[Union[str, Tuple[str, bool], list]] = []
+    """Default sort column if no sorting is applied.
+
+    ???+ example
+        ```python
+        class UserAdmin(ModelView, model=User):
+            column_default_sort = "email"
+        ```
+
+    You can use tuple to control ascending descending order. In following example, items
+    will be sorted in descending order:
+
+    ???+ example
+        ```python
+        class UserAdmin(ModelView, model=User):
+            column_default_sort = ("email", True)
+        ```
+
+    If you want to sort by more than one column, you can pass a list of tuples
+
+    ???+ example
+        ```python
+        class UserAdmin(ModelView, model=User):
+            column_default_sort = [("email", True), ("name", False)]
+        ```
+    """
+
     # Details page
     column_details_list: ClassVar[Sequence[Union[str, InstrumentedAttribute]]] = []
     """List of columns to display in `Detail` page.
@@ -287,7 +361,7 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
 
     ???+ example
         ```python
-        class UserAdmin(ModelAdmin, model=User):
+        class UserAdmin(ModelView, model=User):
             column_details_list = [User.id, User.name, User.mail]
         ```
     """
@@ -300,7 +374,7 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
 
     ???+ example
         ```python
-        class UserAdmin(ModelAdmin, model=User):
+        class UserAdmin(ModelView, model=User):
             column_details_exclude_list = [User.mail]
         ```
     """
@@ -313,7 +387,7 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
 
     ???+ example
         ```python
-        class UserAdmin(ModelAdmin, model=User):
+        class UserAdmin(ModelView, model=User):
             column_formatters_detail = {User.name: lambda m, a: m.name[:10]}
         ```
 
@@ -322,22 +396,11 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
         ```python
         def formatter(model, attribute):
             # `model` is model instance
-            # `attribute` is a Union[Column, ColumnProperty, RelationshipProperty]
+            # `attribute` is a Union[ColumnProperty, RelationshipProperty]
             pass
         ```
     """
 
-    column_labels: ClassVar[Dict[Union[str, InstrumentedAttribute], str]] = {}
-    """A mapping of column labels, used to map column names to new names.
-    Dictionary keys can be string names or SQLAlchemy columns with string values.
-
-    ???+ example
-        ```python
-        class UserAdmin(ModelAdmin, model=User):
-            column_labels = {User.mail: "Email"}
-        ```
-    """
-
     # Templates
     list_template: ClassVar[str] = "list.html"
     """List view template. Default is `list.html`."""
@@ -358,7 +421,7 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
 
     ???+ example
         ```python
-        class UserAdmin(ModelAdmin, model=User):
+        class UserAdmin(ModelView, model=User):
             column_export_list = [User.id, User.name]
         ```
     """
@@ -369,7 +432,7 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
 
     ???+ example
         ```python
-        class UserAdmin(ModelAdmin, model=User):
+        class UserAdmin(ModelView, model=User):
             column_export_exclude_list = [User.id, User.name]
         ```
     """
@@ -395,7 +458,7 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
         class MyForm(Form):
             name = StringField('Name')
 
-        class MyModelAdmin(ModelAdmin, model=User):
+        class MyModelView(ModelView, model=User):
             form = MyForm
         ```
     """
@@ -411,7 +474,7 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
             def do_something(self):
                 pass
 
-        class MyModelAdmin(ModelAdmin, model=User):
+        class MyModelView(ModelView, model=User):
             form_base_class = MyBaseForm
         ```
     """
@@ -424,13 +487,28 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
         ```python
         from wtforms.validators import DataRequired
 
-        class MyModelAdmin(ModelAdmin, model=User):
+        class MyModelView(ModelView, model=User):
             form_args = dict(
                 name=dict(label="User Name", validators=[DataRequired()])
             )
         ```
     """
 
+    form_widget_args: ClassVar[Dict[str, Dict[str, Any]]] = {}
+    """Dictionary of form widget rendering arguments.
+    Use this to customize how widget is rendered without using custom template.
+
+    ???+ example
+        ```python
+        class UserAdmin(ModelView, model=User):
+            form_widget_args = {
+                "email": {
+                    "readonly": True,
+                },
+            }
+        ```
+    """
+
     form_columns: ClassVar[Sequence[Union[str, InstrumentedAttribute]]] = []
     """List of columns to include in the form.
     Columns can either be string names or SQLAlchemy columns.
@@ -440,7 +518,7 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
 
     ???+ example
         ```python
-        class UserAdmin(ModelAdmin, model=User):
+        class UserAdmin(ModelView, model=User):
             form_columns = [User.name, User.mail]
         ```
     """
@@ -451,7 +529,7 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
 
     ???+ example
         ```python
-        class UserAdmin(ModelAdmin, model=User):
+        class UserAdmin(ModelView, model=User):
             form_excluded_columns = [User.id]
         ```
     """
@@ -461,12 +539,84 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
 
     ???+ example
         ```python
-        class UserAdmin(ModelAdmin, model=User):
+        class UserAdmin(ModelView, model=User):
             form_overrides = dict(name=wtf.FileField)
         ```
     """
 
+    form_include_pk: ClassVar[bool] = False
+    """Control if form should include primary key columns or not.
+
+    ???+ example
+        ```python
+        class UserAdmin(ModelView, model=User):
+            form_include_pk = True
+        ```
+    """
+
+    # General options
+    column_labels: ClassVar[Dict[Union[str, InstrumentedAttribute], str]] = {}
+    """A mapping of column labels, used to map column names to new names.
+    Dictionary keys can be string names or SQLAlchemy columns with string values.
+
+    ???+ example
+        ```python
+        class UserAdmin(ModelView, model=User):
+            column_labels = {User.mail: "Email"}
+        ```
+    """
+
+    column_type_formatters: ClassVar[Dict[Type, Callable]] = BASE_FORMATTERS
+    """Dictionary of value type formatters to be used in the list view.
+
+    By default, two types are formatted:
+
+        - None will be displayed as an empty string
+        - bool will be displayed as a checkmark if it is True otherwise as an X.
+
+    If you don’t like the default behavior and don’t want any type formatters applied,
+    just override this property with an empty dictionary:
+
+    ???+ example
+        ```python
+        class UserAdmin(ModelView, model=User):
+            column_type_formatters = dict()
+        ```
+    """
+
+    list_query: ClassVar[Select] = select()
+    """
+    The SQLAlchemy select expression used for the list page which can be customized.
+    By default it will select all objects without any filters.
+
+    ???+ example
+        ```python
+        from sqlalchemy import select
+
+        class UserAdmin(ModelView, model=User):
+            list_query = select(User).filter(User.active == True)
+        ```
+    """
+
+    count_query: ClassVar[Select] = select()
+    """
+    The SQLAlchemy select expression used for the count query which can be customized.
+    By default it will select all objects without any filters.
+
+    ???+ example
+        ```python
+        from sqlalchemy import select
+
+        class UserAdmin(ModelView, model=User):
+            count_query = select(func.count(User.id))
+        ```
+    """
+
     def __init__(self) -> None:
+        self._mapper = inspect(self.model)
+        self._relations = get_relationships(self.model)
+        self._attrs = get_attributes(self.model)
+
         self._column_labels = self.get_column_labels()
         self._column_labels_value_by_key = {
             v: k for k, v in self._column_labels.items()
@@ -478,11 +628,6 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
             for (name, attr) in self._list_attrs
             if isinstance(attr, ColumnProperty)
         ]
-        self._list_relations = [
-            (name, attr)
-            for (name, attr) in self._list_attrs
-            if isinstance(attr, RelationshipProperty)
-        ]
 
         self._details_attrs = self.get_details_columns()
         self._details_columns = [
@@ -490,11 +635,6 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
             for (name, attr) in self._details_attrs
             if isinstance(attr, ColumnProperty)
         ]
-        self._details_relations = [
-            (name, attr)
-            for (name, attr) in self._details_attrs
-            if isinstance(attr, RelationshipProperty)
-        ]
 
         column_formatters = getattr(self, "column_formatters", {})
         self._list_formatters = {
@@ -514,57 +654,29 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
 
         self._search_fields = [
             getattr(self.model, self.get_model_attr(attr).key)
-            for attr in self.column_searchable_list or []
+            for attr in self.column_searchable_list
         ]
 
         self._sort_fields = [
             getattr(self.model, self.get_model_attr(attr).key)
-            for attr in self.column_sortable_list or []
+            for attr in self.column_sortable_list
         ]
 
     def _run_query_sync(self, stmt: ClauseElement) -> Any:
         with self.sessionmaker(expire_on_commit=False) as session:
             result = session.execute(stmt)
-            return result.scalars().all()
+            return result.scalars().unique().all()
 
     async def _run_query(self, stmt: ClauseElement) -> Any:
         if self.async_engine:
             async with self.sessionmaker(expire_on_commit=False) as session:
                 result = await session.execute(stmt)
-                return result.scalars().all()
+                return result.scalars().unique().all()
         else:
             return await anyio.to_thread.run_sync(self._run_query_sync, stmt)
 
-    def _add_object_sync(self, obj: Any) -> None:
-        with self.sessionmaker.begin() as session:
-            session.add(obj)
-
-    def _delete_object_sync(self, obj: Any) -> None:
-        with self.sessionmaker.begin() as session:
-            session.delete(obj)
-
-    def _update_modeL_sync(self, pk: Any, data: Dict[str, Any]) -> None:
-        stmt = select(self.model).where(
-            self.pk_column == self._get_column_python_type(self.pk_column)(pk)
-        )
-        relationships = inspect(self.model).relationships
-
-        with self.sessionmaker.begin() as session:
-            result = session.execute(stmt).scalars().first()
-            for name, value in data.items():
-                if name in relationships and isinstance(value, list):
-                    # Load relationship objects into session
-                    session.add_all(value)
-                setattr(result, name, value)
-
-    def _get_column_python_type(self, column: Column) -> type:
-        try:
-            return column.type.python_type
-        except NotImplementedError:
-            return str
-
     def _url_for_details(self, obj: Any) -> str:
-        pk = getattr(obj, inspect(obj).mapper.primary_key[0].name)
+        pk = getattr(obj, get_primary_key(obj).name)
         return self.url_path_for(
             "admin:details",
             identity=slugify_class_name(obj.__class__.__name__),
@@ -572,7 +684,7 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
         )
 
     def _url_for_edit(self, obj: Any) -> str:
-        pk = getattr(obj, inspect(obj).mapper.primary_key[0].name)
+        pk = getattr(obj, get_primary_key(obj).name)
         return self.url_path_for(
             "admin:edit",
             identity=slugify_class_name(obj.__class__.__name__),
@@ -580,7 +692,7 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
         )
 
     def _url_for_delete(self, obj: Any) -> str:
-        pk = getattr(obj, inspect(obj).mapper.primary_key[0].name)
+        pk = getattr(obj, get_primary_key(obj).name)
         return self.url_path_for(
             "admin:delete",
             identity=slugify_class_name(obj.__class__.__name__),
@@ -589,6 +701,9 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
 
     def _url_for_details_with_attr(self, obj: Any, attr: RelationshipProperty) -> str:
         target = getattr(obj, attr.key)
+        if target is None:
+            return ""
+
         pk = getattr(target, attr.mapper.primary_key[0].name)
         return self.url_path_for(
             "admin:details",
@@ -596,9 +711,26 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
             pk=pk,
         )
 
+    def _get_default_sort(self) -> List[Tuple[str, bool]]:
+        if self.column_default_sort:
+            if isinstance(self.column_default_sort, list):
+                return self.column_default_sort
+            if isinstance(self.column_default_sort, tuple):
+                return [self.column_default_sort]
+            else:
+                return [(self.column_default_sort, False)]
+
+        return [(self.pk_column.name, False)]
+
+    def _default_formatter(self, value: Any) -> Any:
+        if type(value) in self.column_type_formatters:
+            formatter = self.column_type_formatters[type(value)]
+            return formatter(value)
+
+        return value
+
     async def count(self) -> int:
-        stmt = select(func.count(self.pk_column))
-        rows = await self._run_query(stmt)
+        rows = await self._run_query(self.count_query)
         return rows[0]
 
     async def list(
@@ -612,20 +744,24 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
         page_size = min(page_size or self.page_size, max(self.page_size_options))
 
         count = await self.count()
-        stmt = select(self.model).limit(page_size).offset((page - 1) * page_size)
+        stmt = self.list_query.limit(page_size).offset((page - 1) * page_size)
 
-        for _, relation in self._list_relations:
-            stmt = stmt.options(selectinload(relation.key))
+        for relation in self._relations:
+            stmt = stmt.options(joinedload(relation.key))
 
-        sort_field = self.get_model_attr(sort_by) if sort_by else self.pk_column
-        if sort == "desc":
-            stmt = stmt.order_by(desc(sort_field))
+        if sort_by:
+            sort_fields = [(sort_by, sort == "desc")]
         else:
-            stmt = stmt.order_by(asc(sort_field))
+            sort_fields = self._get_default_sort()
+
+        for sort_field, is_desc in sort_fields:
+            if is_desc:
+                stmt = stmt.order_by(desc(sort_field))
+            else:
+                stmt = stmt.order_by(asc(sort_field))
 
         if search:
-            expressions = [attr.ilike(f"%{search}%") for attr in self._search_fields]
-            stmt = stmt.filter(or_(*expressions))
+            stmt = self.search_query(stmt=stmt, term=search)
 
         rows = await self._run_query(stmt)
         pagination = Pagination(
@@ -637,24 +773,23 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
 
         return pagination
 
-    async def get_model_objects(self, limit: int = 0) -> List[Any]:
+    async def get_model_objects(self, limit: Union[int, None] = 0) -> List[Any]:
         # For unlimited rows this should pass None
         limit = None if limit == 0 else limit
         stmt = select(self.model).limit(limit=limit)
 
-        for _, relation in self._list_relations:
-            stmt = stmt.options(selectinload(relation.key))
+        for relation in self._relations:
+            stmt = stmt.options(joinedload(relation.key))
 
         rows = await self._run_query(stmt)
         return rows
 
     async def get_model_by_pk(self, value: Any) -> Any:
-        stmt = select(self.model).where(
-            self.pk_column == self._get_column_python_type(self.pk_column)(value)
-        )
+        pk_value = get_column_python_type(self.pk_column)(value)
+        stmt = select(self.model).where(self.pk_column == pk_value)
 
-        for _, relation in self._details_relations:
-            stmt = stmt.options(selectinload(relation.key))
+        for relation in self._relations:
+            stmt = stmt.options(joinedload(relation.key))
 
         rows = await self._run_query(stmt)
         if rows:
@@ -664,36 +799,40 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
     def get_attr_value(
         self, obj: type, attr: Union[Column, ColumnProperty, RelationshipProperty]
     ) -> Any:
+        result = None
+
         if isinstance(attr, Column):
-            return getattr(obj, attr.name)
+            result = getattr(obj, attr.name)
 
-        value = getattr(obj, attr.key)
-        if isinstance(value, Enum):
-            return value.value
+        if isinstance(attr, (ColumnProperty, RelationshipProperty)):
+            result = getattr(obj, attr.key)
+            result = result.value if isinstance(result, Enum) else result
 
-        return value
+        return result
+
+    def get_list_value(self, obj: type, attr: MODEL_ATTR_TYPE) -> Tuple[Any, Any]:
+        """Get tuple of (value, formatted_value) for the list view."""
+        value = self.get_attr_value(obj, attr)
+        formatted_value = self._default_formatter(value)
 
-    def get_list_value(
-        self, obj: type, attr: Union[Column, ColumnProperty, RelationshipProperty]
-    ) -> Any:
-        """Get instancee values for the list view."""
         formatter = self._list_formatters.get(attr)
         if formatter:
-            return formatter(obj, attr)
-        return self.get_attr_value(obj, attr)
+            formatted_value = formatter(obj, attr)
+        return value, formatted_value
+
+    def get_detail_value(self, obj: type, attr: MODEL_ATTR_TYPE) -> Tuple[Any, Any]:
+        """Get tuple of (value, formatted_value) for the detail view."""
+        value = self.get_attr_value(obj, attr)
+        formatted_value = self._default_formatter(value)
 
-    def get_detail_value(
-        self, obj: type, attr: Union[Column, ColumnProperty, RelationshipProperty]
-    ) -> Any:
-        """Get instancee values for the detail view."""
         formatter = self._detail_formatters.get(attr)
         if formatter:
-            return formatter(obj, attr)
-        return self.get_attr_value(obj, attr)
+            formatted_value = formatter(obj, attr)
+        return value, formatted_value
 
     def get_model_attr(
         self, attr: Union[str, InstrumentedAttribute]
-    ) -> Union[ColumnProperty, RelationshipProperty]:
+    ) -> MODEL_ATTR_TYPE:
         assert isinstance(attr, (str, InstrumentedAttribute))
 
         if isinstance(attr, str):
@@ -708,21 +847,18 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
 
         # Get value by column label
         if key in self._column_labels_value_by_key:
-            return self._column_labels_value_by_key[key]
+            return self._column_labels_value_by_key[str(key)]
 
         raise InvalidColumnError(
             f"Model '{self.model.__name__}' has no attribute '{key}'."
         )
 
-    def get_model_attributes(self) -> List[Column]:
-        return list(inspect(self.model).attrs)
-
     def _build_column_list(
         self,
+        default: List[MODEL_ATTR_TYPE],
         include: Optional[Sequence[Union[str, InstrumentedAttribute]]] = None,
         exclude: Optional[Sequence[Union[str, InstrumentedAttribute]]] = None,
-        default: Callable[[], List[Column]] = None,
-    ) -> List[Tuple[str, Column]]:
+    ) -> List[Tuple[str, MODEL_ATTR_TYPE]]:
         """This function generalizes constructing a list of columns
         for any sequence of inclusions or exclusions.
         """
@@ -730,14 +866,13 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
             attrs = [self.get_model_attr(attr) for attr in include]
         elif exclude:
             exclude_columns = [self.get_model_attr(attr) for attr in exclude]
-            all_attrs = self.get_model_attributes()
-            attrs = list(set(all_attrs) - set(exclude_columns))
+            attrs = list(set(self._attrs) - set(exclude_columns))
         else:
-            attrs = default()
+            attrs = default
 
         return [(self._column_labels.get(attr, attr.key), attr) for attr in attrs]
 
-    def get_list_columns(self) -> List[Tuple[str, Column]]:
+    def get_list_columns(self) -> List[Tuple[str, MODEL_ATTR_TYPE]]:
         """Get list of columns to display in List page."""
 
         column_list = getattr(self, "column_list", None)
@@ -746,10 +881,10 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
         return self._build_column_list(
             include=column_list,
             exclude=column_exclude_list,
-            default=lambda: [getattr(self.model, self.pk_column.name).prop],
+            default=[getattr(self.model, self.pk_column.name).prop],
         )
 
-    def get_details_columns(self) -> List[Tuple[str, Column]]:
+    def get_details_columns(self) -> List[Tuple[str, MODEL_ATTR_TYPE]]:
         """Get list of columns to display in Detail page."""
 
         column_details_list = getattr(self, "column_details_list", None)
@@ -758,36 +893,44 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
         return self._build_column_list(
             include=column_details_list,
             exclude=column_details_exclude_list,
-            default=self.get_model_attributes,
+            default=self._attrs,
         )
 
-    def get_form_columns(self) -> List[Tuple[str, Column]]:
+    def get_form_columns(self) -> List[Tuple[str, MODEL_ATTR_TYPE]]:
         """Get list of columns to display in the form."""
 
         form_columns = getattr(self, "form_columns", None)
         form_excluded_columns = getattr(self, "form_excluded_columns", None)
 
+        columns = list(self._mapper.columns)
+        default = []
+        for attr in self._attrs:
+            if attr in self._relations:
+                default.append(attr)
+            if attr in columns:
+                default.append(attr)
+
         return self._build_column_list(
             include=form_columns,
             exclude=form_excluded_columns,
-            default=self.get_model_attributes,
+            default=self._attrs,
         )
 
-    def get_export_columns(self) -> List[Tuple[str, Column]]:
+    def get_export_columns(self) -> List[Tuple[str, MODEL_ATTR_TYPE]]:
         """Get list of columns to export."""
 
         columns = getattr(self, "column_export_list", None)
         excluded_columns = getattr(self, "column_export_exclude_list", None)
-        if not columns and not excluded_columns:
-            return self.get_list_columns()
 
         return self._build_column_list(
             include=columns,
             exclude=excluded_columns,
-            default=lambda: self._list_columns,
+            default=[item[1] for item in self._list_attrs],
         )
 
-    def get_column_labels(self) -> Dict[Column, str]:
+    def get_column_labels(
+        self,
+    ) -> Dict[MODEL_ATTR_TYPE, str]:
         return {
             self.get_model_attr(column_label): value
             for column_label, value in self.column_labels.items()
@@ -795,38 +938,21 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
 
     async def delete_model(self, obj: Any) -> None:
         if self.async_engine:
-            async with self.sessionmaker.begin() as session:
-                await session.delete(obj)
+            await delete_async(self, obj)
         else:
-            await anyio.to_thread.run_sync(self._delete_object_sync, obj)
+            await anyio.to_thread.run_sync(delete_sync, self, obj)
 
-    async def insert_model(self, obj: type) -> Any:
+    async def insert_model(self, data: dict) -> None:
         if self.async_engine:
-            async with self.sessionmaker.begin() as session:
-                session.add(obj)
+            await insert_async(self, data)
         else:
-            await anyio.to_thread.run_sync(self._add_object_sync, obj)
+            await anyio.to_thread.run_sync(insert_sync, self, data)
 
     async def update_model(self, pk: Any, data: Dict[str, Any]) -> None:
         if self.async_engine:
-            stmt = select(self.model).where(
-                self.pk_column == self._get_column_python_type(self.pk_column)(pk)
-            )
-            relationships = inspect(self.model).relationships
-
-            for name in relationships.keys():
-                stmt = stmt.options(selectinload(name))
-
-            async with self.sessionmaker.begin() as session:
-                result = await session.execute(stmt)
-                result = result.scalars().first()
-                for name, value in data.items():
-                    if name in relationships and isinstance(value, list):
-                        # Load relationship objects into session
-                        session.add_all(value)
-                    setattr(result, name, value)
+            await update_async(self, pk, data)
         else:
-            await anyio.to_thread.run_sync(self._update_modeL_sync, pk, data)
+            await anyio.to_thread.run_sync(update_sync, self, pk, data)
 
     async def scaffold_form(self) -> Type[Form]:
         if self.form is not None:
@@ -834,11 +960,13 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
         return await get_model_form(
             model=self.model,
             engine=self.engine,
-            only=[i[1].key for i in self._form_attrs],
+            only=[i[1].key or i[1].name for i in self._form_attrs],
             column_labels={k.key: v for k, v in self._column_labels.items()},
             form_args=self.form_args,
+            form_widget_args=self.form_widget_args,
             form_class=self.form_base_class,
             form_overrides=self.form_overrides,
+            form_include_pk=self.form_include_pk,
         )
 
     def search_placeholder(self) -> str:
@@ -846,7 +974,7 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
 
         ???+ example
             ```python
-            class UserAdmin(ModelAdmin, model=User):
+            class UserAdmin(ModelView, model=User):
                 column_labels = dict(name="Name", email="Email")
                 column_searchable_list = [User.name, User.email]
 
@@ -862,6 +990,17 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
         ]
         return ", ".join(field_names)
 
+    def search_query(self, stmt: Select, term: str) -> Select:
+        """Specify the search query given the SQLAlchemy statement and term to search for.
+        It can be used for doing more complex queries like JSON objects. For example:
+
+        ```py
+        return stmt.filter(MyModel.name == term)
+        ```
+        """
+        expressions = [attr.ilike(f"%{term}%") for attr in self._search_fields]
+        return stmt.filter(or_(*expressions))
+
     def get_export_name(self, export_type: str) -> str:
         """The file name when exporting."""
         filename = f"{self.name}_{time.strftime('%Y-%m-%d_%H-%M-%S')}.{export_type}"
@@ -881,15 +1020,13 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
         self,
         data: List[Any],
     ) -> StreamingResponse:
-        def generate(writer: Writer) -> Generator[List[str], None, None]:
+        def generate(writer: Writer) -> Generator[Any, None, None]:
             # Append the column titles at the beginning
             titles = [c[0] for c in self._export_attrs]
             yield writer.writerow(titles)
 
             for row in data:
-                vals = [
-                    as_str(self.get_attr_value(row, c[1])) for c in self._export_attrs
-                ]
+                vals = [str(self.get_attr_value(row, c[1])) for c in self._export_attrs]
                 yield writer.writerow(vals)
 
         # `get_export_name` can be subclassed.
@@ -901,3 +1038,14 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
             media_type="text/csv",
             headers={"Content-Disposition": f"attachment;filename={filename}"},
         )
+
+
+class ModelAdmin(ModelView):
+    def __init__(self) -> None:  # pragma: no cover
+        import warnings
+
+        warnings.warn(
+            "The class `ModelAdmin` is deprectated, please use `ModelView instead.`",
+            DeprecationWarning,
+        )
+        super().__init__()
